/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * Bump map plug-in --- emboss an image by using another image as a bump map
 * Copyright (C) 1997 Federico Mena Quintero
 * federico@nuclecu.unam.mx
 * Copyright (C) 1997 Jens Lautenbacher
 * jens@lemming0.lem.uni-karlsruhe.de
 *
 */
/**********************************************************************
  for layman modified by TANE
  $Id: bumpmap.c,v 1.3 1998/10/31 08:50:15 TANE Exp $
**********************************************************************/
#include	"common.h"
#include	"bumpmap.h"
#include	"funcs.h"
#include	<math.h>
#include	"gimplib.h"

/***** Magic numbers *****/

#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif /* M_PI */

#define max(a,b)	(((a) >= (b)) ? (a) : (b))
#define min(a,b)	(((a) <= (b)) ? (a) : (b))

#define PLUG_IN_NAME    "plug_in_bump_map"
#define PLUG_IN_VERSION "August 1997, 2.04"

#define PREVIEW_SIZE 128
#define SCALE_WIDTH  200
#define ENTRY_WIDTH  60

#define CHECK_SIZE  8
#define CHECK_DARK  ((int) (1.0 / 3.0 * 255))
#define CHECK_LIGHT ((int) (2.0 / 3.0 * 255))


/***** Types *****/

typedef struct {
	gint32  bumpmap_id;
	gdouble azimuth;
	gdouble elevation;
	gint    depth;
	gint    xofs;
	gint    yofs;
	gint    waterlevel;
	gint    ambient;
	gint    compensate;
	gint    invert;
	gint    type;
} bumpmap_vals_t;

typedef struct {
	int    lx, ly;       /* X and Y components of light vector */
	int    nz2, nzlz;    /* nz^2, nz*lz */
	int    background;   /* Shade for vertical normals */
	double compensation; /* Background compensation */
	guchar lut[256];     /* Look-up table for modes */
} bumpmap_params_t;

/***** Prototypes *****/

void bumpmap_init_params(bumpmap_params_t *params);
void bumpmap_row(guchar           *src_row,
			guchar           *dest_row,
			int               width,
			int               bpp,
			int               has_alpha,
			guchar           *bm_row1,
			guchar           *bm_row2,
			guchar           *bm_row3,
			int               bm_width,
			int               bm_xofs,
			bumpmap_params_t *params);
void bumpmap_convert_row(guchar *row, int width, int bpp, int has_alpha, guchar *lut);

/***** Variables *****/

static bumpmap_vals_t bmvals = {
	-1,     /* bumpmap_id */
	135.0,  /* azimuth */
	45.0,   /* elevation */
	3,      /* depth */
	0,      /* xofs */
	0,      /* yofs */
	0,      /* waterlevel */
	0,      /* ambient */
	FALSE,  /* compensate */
	FALSE,  /* invert */
	LINEAR  /* type */
}; /* bmvals */

//static char *map_types[] = {
//	"Linear map",
//	"Spherical map",
//	"Sinuosidal map"
//}; /* map_types */

//GDrawable *drawable = NULL;

int sel_x1, sel_y1, sel_x2, sel_y2;
int sel_width, sel_height;
int img_bpp, img_has_alpha;


/***** Functions *****/

int	bumpmap( struct VGRAM_STRUCT *vg_pt, struct BUMPMAP_EXEC_STRUCT *exec_info )
{
	bumpmap_params_t  params;
//	GDrawable        *bm_drawable;
//	int              *bm_drawable;
//	GPixelRgn         src_rgn, dest_rgn, bm_rgn;
//	VGRAM_STRUCT      *src_rgn, *dest_rgn, *bm_rgn;
	int               bm_width, bm_height, bm_bpp, bm_has_alpha;
	int               yofs1, yofs2, yofs3;
	guchar           *bm_row1, *bm_row2, *bm_row3, *bm_tmprow;
	guchar           *src_row, *dest_row;
	int               y;
	int				  i;

	/* Get image information */

	sel_width = vg_pt->width ;
	img_bpp   = 3 ; //sizeof( struct PIX_STRUCT ) ;

	bm_width     = vg_pt->width ;
	bm_height    = vg_pt->height ;
	bm_bpp       = 3 ; //sizeof( struct PIX_STRUCT ) ;
//	bm_has_alpha = gimp_drawable_has_alpha(bm_drawable->id);

	bm_has_alpha  = NO ;
	img_has_alpha = NO ;
	sel_width  = vg_pt->width ;
	sel_height = 1 ;
	sel_x1 = 0 ;
	sel_y1 = 0 ;
	sel_x2 = vg_pt->width-1 ;
	sel_y2 = vg_pt->height-1 ;

	bmvals.type      = exec_info->type ;
	bmvals.azimuth   = exec_info->azimuth ;
	bmvals.elevation = exec_info->elevation ;
	bmvals.depth     = exec_info->depth ;

	/* Initialize offsets */

	yofs2 = bmvals.yofs % bm_height;
	yofs1 = (yofs2 + bm_height - 1) % bm_height;
	yofs3 = (yofs2 + 1) % bm_height;

	/* Initialize row buffers */

	bm_row1 = malloc(bm_width * (bm_bpp) * sizeof(guchar));
	bm_row2 = malloc(bm_width * (bm_bpp) * sizeof(guchar));
	bm_row3 = malloc(bm_width * (bm_bpp) * sizeof(guchar));

	src_row  = malloc(sel_width * (img_bpp) * sizeof(guchar));
	dest_row = malloc(sel_width * (img_bpp) * sizeof(guchar));

	/* Initialize pixel regions */

//	gimp_pixel_rgn_init(src_rgn, drawable, sel_x1, sel_y1, sel_width, sel_height, FALSE, FALSE);
//	gimp_pixel_rgn_init(dest_rgn, drawable, sel_x1, sel_y1, sel_width, sel_height, TRUE, TRUE);
//	gimp_pixel_rgn_init(bm_rgn, bm_drawable, 0, 0, bm_width, bm_height, FALSE, FALSE);


	/* Bumpmap */

	bumpmap_init_params(&params);

//	gimp_pixel_rgn_get_row(bm_rgn, bm_row1, 0, yofs1, bm_width);
//	gimp_pixel_rgn_get_row(bm_rgn, bm_row2, 0, yofs2, bm_width);
//	gimp_pixel_rgn_get_row(bm_rgn, bm_row3, 0, yofs3, bm_width);

	for( i=0 ; i< bm_width ; i++ ){
		bm_row1[ i*bm_bpp+0 ] = vg_pt->pix[i+yofs1*vg_pt->width].r ;
		bm_row1[ i*bm_bpp+1 ] = vg_pt->pix[i+yofs1*vg_pt->width].g ;
		bm_row1[ i*bm_bpp+2 ] = vg_pt->pix[i+yofs1*vg_pt->width].b ;
		bm_row2[ i*bm_bpp+0 ] = vg_pt->pix[i+yofs2*vg_pt->width].r ;
		bm_row2[ i*bm_bpp+1 ] = vg_pt->pix[i+yofs2*vg_pt->width].g ;
		bm_row2[ i*bm_bpp+2 ] = vg_pt->pix[i+yofs2*vg_pt->width].b ;
		bm_row3[ i*bm_bpp+0 ] = vg_pt->pix[i+yofs3*vg_pt->width].r ;
		bm_row3[ i*bm_bpp+1 ] = vg_pt->pix[i+yofs3*vg_pt->width].g ;
		bm_row3[ i*bm_bpp+2 ] = vg_pt->pix[i+yofs3*vg_pt->width].b ;
	}

	bumpmap_convert_row(bm_row1, bm_width, bm_bpp, bm_has_alpha, params.lut);
	bumpmap_convert_row(bm_row2, bm_width, bm_bpp, bm_has_alpha, params.lut);
	bumpmap_convert_row(bm_row3, bm_width, bm_bpp, bm_has_alpha, params.lut);

	for (y = sel_y1; y <= sel_y2; y++) {
//		gimp_pixel_rgn_get_row(&src_rgn, src_row, sel_x1, y, sel_width);
		for( i=0 ; i< sel_width ; i++ ){
			src_row[i*img_bpp+0] = vg_pt->pix[i+y*vg_pt->width].r ;
			src_row[i*img_bpp+1] = vg_pt->pix[i+y*vg_pt->width].g ;
			src_row[i*img_bpp+2] = vg_pt->pix[i+y*vg_pt->width].b ;
		}

		bumpmap_row(src_row, dest_row, sel_width, img_bpp, img_has_alpha,
			    bm_row1, bm_row2, bm_row3, bm_width, bmvals.xofs, &params);

//		gimp_pixel_rgn_set_row(&dest_rgn, dest_row, sel_x1, y, sel_width);
		for( i=0 ; i< sel_width ; i++ ){
			vg_pt->pix[i+y*vg_pt->width].r = dest_row[i*img_bpp+0] ;
			vg_pt->pix[i+y*vg_pt->width].g = dest_row[i*img_bpp+1] ;
			vg_pt->pix[i+y*vg_pt->width].b = dest_row[i*img_bpp+2] ;
		}

		/* Next line */

		bm_tmprow = bm_row1;
		bm_row1   = bm_row2;
		bm_row2   = bm_row3;
		bm_row3   = bm_tmprow;
		
		if (++yofs3 == bm_height)
			yofs3 = 0;

//		gimp_pixel_rgn_get_row(&bm_rgn, bm_row3, 0, yofs3, bm_width);

		for( i=0 ; i< bm_width ; i++ ){
			bm_row3[ i*bm_bpp+0 ] = vg_pt->pix[i+yofs3*vg_pt->width].r ;
			bm_row3[ i*bm_bpp+1 ] = vg_pt->pix[i+yofs3*vg_pt->width].g ;
			bm_row3[ i*bm_bpp+2 ] = vg_pt->pix[i+yofs3*vg_pt->width].b ;
		}
		bumpmap_convert_row(bm_row3, bm_width, bm_bpp, bm_has_alpha, params.lut);

	} /* for */

	/* Done */

	free(bm_row1);
	free(bm_row2);
	free(bm_row3);
	free(src_row);
	free(dest_row);

	return( NORMAL_EXIT ) ;

} /* bumpmap */


/*****/

void bumpmap_init_params(bumpmap_params_t *params)
{
	double azimuth;
	double elevation;
	int    lz, nz;
	int    i;
	double n;

	/* Convert to radians */

	azimuth   = M_PI * bmvals.azimuth / 180.0;
	elevation = M_PI * bmvals.elevation / 180.0;

	/* Calculate the light vector */

	params->lx = cos(azimuth) * cos(elevation) * 255.0;
	params->ly = sin(azimuth) * cos(elevation) * 255.0;
	lz         = sin(elevation) * 255.0;

	/* Calculate constant Z component of surface normal */

	nz           = (6 * 255) / bmvals.depth;
	params->nz2  = nz * nz;
	params->nzlz = nz * lz;

	/* Optimize for vertical normals */

	params->background = lz;

	/* Calculate darkness compensation factor */

	params->compensation = sin(elevation);

	/* Create look-up table for map type */

//	fprintf(stderr,"type=%d\n",bmvals.type) ;
 	for(i = 0; i < 256; i++) {
		switch( bmvals.type ) {
			case SPHERICAL:
				n = i / 255.0 - 1.0;
				params->lut[i] = (int) (255.0 * sqrt(1.0 - n * n) + 0.5);
				break;

			case SINUOSIDAL:
				n = i / 255.0;
				params->lut[i] = (int) (255.0 * (sin((-M_PI / 2.0) + M_PI * n) + 1.0) /
							2.0 + 0.5);
				break;

			case LINEAR:
			default:
				params->lut[i] = i;
		} /* switch */

		if( bmvals.invert )
			params->lut[i] = 255 - params->lut[i];
	} /* for */
} /* bumpmap_init_params */


/*****/

void
bumpmap_row(guchar           *src_row,
	    guchar           *dest_row,
	    int               width,
	    int               bpp,
	    int               has_alpha,
	    guchar           *bm_row1,
	    guchar           *bm_row2,
	    guchar           *bm_row3,
	    int               bm_width,
	    int               bm_xofs,
	    bumpmap_params_t *params)
{
	guchar *src, *dest;
	int     xofs1, xofs2, xofs3;
	int     shade;
	int     ndotl;
	int     nx, ny;
	int     x, k;
	int     pbpp;
	int     result;

	if (has_alpha)
		pbpp = bpp - 1;
	else
		pbpp = bpp;

	if (bm_xofs < 0)
		xofs2 = bm_width - (-bm_xofs % bm_width);
	else
		xofs2 = bm_xofs % bm_width;

	xofs1 = (xofs2 + bm_width - 1) % bm_width;
	xofs3 = (xofs2 + 1) % bm_width;

	src  = src_row;
	dest = dest_row;

	for (x = 0; x < width; x++) {
		/* Calculate surface normal from bump map */

		nx = (bm_row1[xofs1] + bm_row2[xofs1] + bm_row3[xofs1] -
		      bm_row1[xofs3] - bm_row2[xofs3] - bm_row3[xofs3]);
		ny = (bm_row3[xofs1] + bm_row3[xofs2] + bm_row3[xofs3] -
		      bm_row1[xofs1] - bm_row1[xofs2] - bm_row1[xofs3]);

		/* Shade */

		if ((nx == 0) && (ny == 0))
			shade = params->background;
		else {
			ndotl = nx * params->lx + ny * params->ly + params->nzlz;

			if (ndotl < 0)
				shade = params->compensation * bmvals.ambient;
			else {
				shade = ndotl / sqrt(nx * nx + ny * ny + params->nz2);

				shade = shade + max(0, (255 * params->compensation - shade)) *
					        bmvals.ambient / 255;
			} /* else */
		} /* else */

		/* Paint */

		if (bmvals.compensate)
			for (k = pbpp; k; k--) {
				result  = (*src++ * shade) / (params->compensation * 255);
				*dest++ = min(255, result);
			} /* for */
		else
			for (k = pbpp; k; k--)
				*dest++ = *src++ * shade / 255;
		
		if (has_alpha)
			*dest++ = *src++;

		/* Next pixel */

		if (++xofs1 == bm_width)
			xofs1 = 0;

		if (++xofs2 == bm_width)
			xofs2 = 0;

		if (++xofs3 == bm_width)
			xofs3 = 0;
	} /* for */
} /* bumpmap_row */


/*****/

void
bumpmap_convert_row(guchar *row, int width, int bpp, int has_alpha, guchar *lut)
{
	guchar *p;

	p = row;

	has_alpha = has_alpha ? 1 : 0;

	if (bpp >= 3)
		for (; width; width--) {
			if (has_alpha)
				*p++ = lut[(int) (bmvals.waterlevel +
						  (((int)
						    (0.30 * row[0] + 0.59 * row[1] + 0.11 * row[2] + 0.5) -
						    bmvals.waterlevel) *
						   row[3]) / 255.0)];
			else
				*p++ = lut[(int) (0.30 * row[0] + 0.59 * row[1] + 0.11 * row[2] + 0.5)];

			row += 3 + has_alpha;
		} /* for */
	else
		for (; width; width--) {
			if (has_alpha)
				*p++ = lut[bmvals.waterlevel +
					   ((row[0] - bmvals.waterlevel) * row[1]) / 255];
			else
				*p++ = lut[*row];

			row += 1 + has_alpha;
		} /* for */
} /* bumpmap_convert_row */


